import boto3
import pandas as pd
import io

class S3FileHandler:
    """
    Performs read, write, move, copy actions on files in S3 buckets operations on various files
    """

    s3 = boto3.resource('s3')

    def __init__(self, source, key):
        self.source = source
        self.key = key
        self.filetype = self.extract_filetype()

    def set_source(self, source):
        self.source = source

    def set_key(self, key):
        self.key = key

    def extract_filetype(self):
        try:
            filetype = self.key.split('/')[-1].split('.')[-1]
            return filetype
        except:
            return None

    def copy(self, target, new_key=None):
        copy_source = {
            'Bucket': self.source,
            'Key': self.key
        }

        if new_key == None:
            new_key = self.key

        self.s3.meta.client.copy(CopySource=copy_source, Bucket=target, Key=new_key)
        print('Copied file to {}:{}'.format(target, self.key))

    def delete(self):
        self.s3.meta.client.delete_object(Bucket=self.source, Key=self.key)
        print('Deleted file {}:{}'.format(self.source, self.key))

    def move(self, target):
        self.s3.meta.client.copy(self.source, target, self.key)
        self.delete()
        self.source = target
        print('Moved file to {}:{}'.format(target, self.key))

    def get_object(self):
        obj = self.s3.meta.client.get_object(Bucket=self.source, Key=self.key)
        return obj

    def put_object(self, obj, target, key):
        self.s3.meta.client.put_object(Body=obj, Bucket=target, Key=key)
        print('Put file {}:{}'.format(target, key))

    def read_csv(self, sep=',', dtype=None):
        if (self.filetype == 'csv'):
            obj = self.get_object()
            bytes_obj = io.BytesIO(obj['Body'].read())
            self.dataframe = pd.read_csv(bytes_obj, sep=sep, dtype=dtype)
            return self.dataframe
        else:
            raise AttributeError('Not a .csv file')

    def read_excel(self, dtype=None):
        if (self.filetype in ['xlsx', 'xls']):
            obj = self.get_object()
            bytes_obj = io.BytesIO(obj['Body'].read())
            self.dataframe = pd.read_excel(bytes_obj, dtype=dtype)
            return self.dataframe
        else:
            raise AttributeError('Not a .xlsx file')

    def read_json(self, dtype=None):
        if (self.filetype == 'json'):
            obj = self.get_object()
            str_obj = obj['Body'].read().decode('UTF-8')
            self.dataframe = pd.read_json(str_obj, dtype=dtype)
            return self.dataframe
        else:
            raise AttributeError('Not a .json file')

    def read_file(self, dtype=None, sep=','):
        if self.filetype == 'json':
            self.read_json(dtype=dtype)
        elif self.filetype == 'csv':
            self.read_csv(dtype=dtype, sep=sep)
        elif self.filetype in ['xlsx', 'xls']:
            self.read_excel(dtype=dtype)
        else:
            raise AttributeError("Filetype '.{}' not recognized".format(self.filetype))

    def to_dict(self, orient='records'):
        try:
            dict_ = self.dataframe.to_dict(orient=orient, into=NativeDict)
            return dict_
        except:
            raise AttributeError('Dataframe not present')

    def to_json(self, orient='records'):
        try:
            json = self.dataframe.to_json(orient=orient)
            return json
        except:
            raise AttributeError('Dataframe not present')

    def put_csv(self, target, key, sep=','):
        try:
            csv_file = self.to_csv(sep=sep)
            self.put_object(obj=csv_file, target=target, key=key)
        except:
            raise AttributeError('Dataframe not present')

    def put_excel(self, target, key, excel_writer=None):
        try:
            excel_file = self.to_excel(excel_writer=excel_writer)
            self.put_object(obj=excel_file, target=target, key=key)
        except:
            raise AttributeError('Dataframe not present')

    def put_json(self, target, key, orient='records'):
        try:
            json_file = self.to_json(orient=orient)
            self.put_object(obj=json_file, target=target, key=key)
        except:
            raise AttributeError('Dataframe not present')


class NativeDict(dict):
    """
        Helper class to ensure that only native types are in the dicts produced by
        :func:`to_dict() <pandas.DataFrame.to_dict>`

        .. note::

            Needed until `#21256 <https://github.com/pandas-dev/pandas/issues/21256>`_ is resolved.
    """
    def __init__(self, *args, **kwargs):
        super().__init__(((k, self.convert_if_needed(v)) for row in args for k, v in row), **kwargs)

    @staticmethod
    def convert_if_needed(value):
        """
            Converts `value` to native python type.

            .. warning::

                Only :class:`Timestamp <pandas.Timestamp>` and numpy :class:`dtypes <numpy.dtype>` are converted.
        """
        if pd.isnull(value):
            return None
        if isinstance(value, pd.Timestamp):
            return value.to_pydatetime()
        if hasattr(value, 'dtype'):
            mapper = {'i': int, 'u': int, 'f': float}
            _type = mapper.get(value.dtype.kind, lambda x: x)
            return _type(value)
        return value



import sys
from sqlalchemy import create_engine
from sqlalchemy.ext.automap import automap_base
from sqlalchemy.orm import Session
from sqlalchemy.types import (Integer, String, DateTime, TypeDecorator,
                              JSON, TEXT, Enum, TIMESTAMP, Float, Unicode,
                              Numeric)
from sqlalchemy.exc import IntegrityError
from sqlalchemy.orm.exc import FlushError
import hashlib
import pandas as pd


class SQLAlchemyInserter:
    """
    Takes records (list of dictionaries) and inserts them into a database using a list of targets
    """
    def __init__(self, database_url):
        self.database_url = database_url
        self.engine = self.set_engine()
        self.connection = self.engine.connect()
        self.session = self.make_database_session()
        self.upload_queue = []

    def set_engine(self):
        """
        Set engine
        """
        return create_engine(self.database_url)

    def make_database_session(self):
        """
        Make database session from sqlalchemy engine
        """
        return Session(bind=self.engine)

    def add_to_upload_queue(self, records, table, orm_model=None):
        if orm_model == None:
            orm_model = self.infer_orm_model(table)

        upload_object = {
            'records_hash' : self.compute_hash(records),
            'records_count' : self.count_records(records),
            'records_size' : self.get_size(records),
            'data' : records,
            'table' : table,
            'orm_model' : orm_model
        }
        self.upload_queue.append(upload_object)

    def infer_orm_model(self, table):
        """
        Infers SQLAlchemy ORM model from database schema of table
        """
        Base = automap_base()
        Base.prepare(self.engine, reflect=True)
        orm_model = Base.classes[table]
        return orm_model

    def get_orm_dtypes(self, orm_model):
        """
        Takes SQLAlchemy ORM model and returns dictionary of python datatypes { column_name : column_type }
        """
        orm_dtypes = { cName : self.convert_orm_dtypes(cValue.type) for cName, cValue in orm_model.__table__.columns.items()}
        return orm_dtypes

    def convert_orm_dtypes(self, SQLAlchemyDatatype):
        """
        Converts SQLAlchemy datatype to pandas datatype
        """
        if isinstance(SQLAlchemyDatatype, Integer):
            return "int64"
        elif isinstance(SQLAlchemyDatatype, DateTime):
            return "datetime64[ns]"
        elif isinstance(SQLAlchemyDatatype, Float):
            return "float64"
        elif isinstance(SQLAlchemyDatatype, Unicode):
            return "unicode"
        elif isinstance(SQLAlchemyDatatype, String):
            return "unicode"
        else:
            raise ValueError("Datatype {} not recognized".format(SQLAlchemyDatatype))

    def get_orm_columns(self, orm_model):
        """
        Retrieve columns from ORM model
        """
        return list(orm_model.__table__.columns.keys())

    def insert_or_update_to_database(self, record, orm_model):
        instance = orm_model(**record)
        try:
            print(f"Adding instance to table, table {orm_model.__tablename__}")
            self.session.add(instance)
            self.session.commit()
        except (IntegrityError, FlushError) as e:
            print(f"Key already exists, rolling back and checking chronology of events")
            self.session.rollback()
            try:
                primary_key = instance.id
                if primary_key is None:
                    print(f"Record with ID None: {record}")
                print(f"ID {primary_key} was already present in table {orm_model.__tablename__}, checking chronology")
                row = self.session.query(orm_model).filter(orm_model.id == primary_key).first()
                try:
                    old_event_time = pd.to_datetime(row.modified).replace(tzinfo=None)
                    current_event_time = pd.to_datetime(instance.modified).replace(tzinfo=None)
                    print(f"Current event was modified at {instance.modified}, old event was modified at {row.modified}")
                except Exception:
                    print(f"Event {orm_model.__tablename__} does not have a modified field, using created instead.")
                    old_event_time = pd.to_datetime(row.created).replace(tzinfo=None)
                    current_event_time = pd.to_datetime(instance.created).replace(tzinfo=None)
                    print(f"Current event was created at {instance.created}, old event was created at {row.created}")
                newer = current_event_time > old_event_time
                self.session.query(orm_model).filter(orm_model.id == primary_key).update(record)
                self.session.commit()
                if newer:
                    print("Event is newer than previously inserted data, the data have been updated.")
                else: 
                    print(f"event was older than previously inserted data, event has been discarded")
            except Exception as e:
                print(record)
                print("Instance is missing fields")
                self.session.rollback()
                raise(e)
                        
    def insert_or_update_all(self):
        total_len = len(self.upload_queue)
        counter = 0
        for upload_object in self.upload_queue:
            counter = counter + 1 
            print(f"Uploading object {counter} out of {total_len}")
            for record in upload_object['data']:
                self.insert_or_update_to_database(
                    record=record,
                    orm_model=upload_object['orm_model']
                )
            print("All objects in queue have been processed.")


    def push_to_database(self, records, orm_model, table, record_count):
        """
        Push records to database
        """
        try:
            self.session.bulk_insert_mappings(orm_model, records)
            self.session.commit()
            print('Inserted {} records of size into table "{}"'.format(record_count, table))
        except:
            self.session.rollback()
            raise
        finally:
            self.session.close()

    def push_all_to_database(self):
        """
        Push all records of SQLAlchemyInsert instance to database
        """
        for upload_object in self.upload_queue:
            self.push_to_database(
                    records = upload_object['data'],
                    orm_model = upload_object['orm_model'],
                    table = upload_object['table'],
                    record_count = upload_object['records_count']
            )

    def sort_upload_queue(self, ascending=True, key='record_count'):
        if ascending == True:
            reverse = False
        elif ascending == False:
            reverse = True

        sorted_queue = sorted(self.upload_queue, key=lambda k: k[key], reverse=reverse)
        self.upload_queue = sorted_queue

    def count_records(self, records):
        return len(records)

    def get_size(self, obj, seen=None, unit=None):
        """
        Recursively finds size of objects and returns bytes
        """
        size = sys.getsizeof(obj)
        if seen is None:
            seen = set()
        obj_id = id(obj)
        if obj_id in seen:
            return 0
        # Important mark as seen *before* entering recursion to gracefully handle
        # self-referential objects
        seen.add(obj_id)
        if isinstance(obj, dict):
            size += sum([self.get_size(v, seen) for v in obj.values()])
            size += sum([self.get_size(k, seen) for k in obj.keys()])
        elif hasattr(obj, '__dict__'):
            size += self.get_size(obj.__dict__, seen)
        elif hasattr(obj, '__iter__') and not isinstance(obj, (str, bytes, bytearray)):
            size += sum([self.get_size(i, seen) for i in obj])

        return self.transform_byte_size(size, unit=unit)

    def transform_byte_size(self, byte_size, unit=None):
        byte_size = float(byte_size)
        if unit == None:
            return byte_size
        elif unit == 'kb':
            return byte_size / 10**3
        elif unit == 'Mb':
            return byte_size / 10**6
        elif unit == 'Gb':
            return byte_size / 10**9
        elif unit == 'Tb':
            return byte_size / 10**12
        else:
            raise ValueError("Unit unknown to function")

    def compute_hash(self, obj):
        """
        Compute Sha1 hexadecimal hash string from object
        """
        hash_object = hashlib.sha1(str(obj).encode())
        return hash_object.hexdigest()

    @property
    def queue_size(self, unit=None):
        size = 0.0
        for upload_object in self.upload_queue:
            size += self.transform_byte_size(upload_object['records_size'], unit=unit)
        return size

    @property
    def queue_tables(self):
        return [upload_object['table'] for upload_object in self.upload_queue]

    @property
    def queue_length(self):
        return len(self.upload_queue)

class PyMongoInserter(SQLAlchemyInserter):
    def push_bulk_to_database(self):
        return True   



import setuptools

with open("README.md", "r") as fh:
    long_description = fh.read()

setuptools.setup(
    name="datalake_vqd_utils", # Replace with your own username
    version="0.0.21",
    author="Casper van Houten",
    author_email="casper.vanhouten@viqtordavis.com",
    description="Package containing tools for cloud-based data lakes",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/pypa/sampleproject",
    packages=setuptools.find_packages(),
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    install_requires=[
    'boto3>=1.9.222',
    'sqlalchemy>=1.3.13',
    'pymysql>=0.9.3',
    'pandas>=0.23.4'
    ],
    python_requires='>=3.6',
)
